/*-------------------------------------*
 * File: runner.c                      *
 *-------------------------------------*
 * RUNNER version 1.0                  *
 * Written by Laurent Debert, Jan 2002 *
 *-------------------------------------*/

#include <rex/rex.h>
#include "bitmaps.c"
#include "levels.c"

/*--------------------------
    Constants & Parameters
  --------------------------*/

#define DEBUG	0	// 0 = REX6000 version, 1 = emulator version

#if DEBUG
	//--- emulator parameters
	#define JUMP_SPEED		100
	#define LASER_SPEED		1100
	#define BLINK_SPEED		250
	#define INCREASE_SPEED	10

	int DsTrace(char *s, ...) 
	{ 
		output8(0xfe, 0); 
	}
#else
	//-- realtime parameters
	#define JUMP_SPEED		14
	#define LASER_SPEED		160
	#define BLINK_SPEED		30
	#define INCREASE_SPEED	2

	#define DsTrace(x)		
#endif

//-- game constants
#define NLEVELS			40
#define LVL_DATA_SIZE	35
#define POINT_HEART		5
#define POINT_LEVEL		100
#define NPLAYERS		4

//-- elements
#define ZWALL			0x01
#define ZLADER			0x02
#define ZLADER2			0x04
#define ZHEART			0x08

//-- events
#define E_ALEFT			49
#define E_ARIGHT		51
#define E_AUP			30
#define E_ADOWN			70
#define E_RESTART		71
#define E_PREV			72
#define E_NEXT			73
#define E_AUPL			80
#define E_AUPR			81

/*-------------
    Variables
  -------------*/

int  curpos;		// current player position in the screen
int  curlevel;		// current level
int  score;			// current score
int  laserspeed;	// current speed of the lasers
int  maxspeed;		// maximum speed of the lasers
char start;			// =1 if game is started otherwise =0
char onlader;		// =1 if player is currently climbing a lader otherwise =0
char players;		// number of players left
char hearts;		// numbers of hearts still to pick up
char laserpos;		// current position of the lasers
char *plevel;		// pointer to the current level data array
char *manstate;		// current state of the player
char screen[20*13]; // screen (array of cells: 20 columns * 13 rows)
int  difficulty;	// 1=Medium, 2=Easy

/*-------------
    Functions
  -------------*/

//-- Draw any bitmap at cell position (player, blank, heart, wall or lader)
DrawCell(pos,bmp) int pos; char *bmp;
{
   int x,y;
   x = 4 + (pos % 20)*8;
   y = 4 + (pos / 20)*8;
   if (bmp) DsDisplayBitmapDraw(x,y,bmp,0);
   else DsDisplayBlockClear(x,y,8,8);
}

//-- RestoreCell(): Restore a cell of the level (i.e. blank, wall or lader)
RestoreCell(pos) int pos;
{
	switch(screen[pos])
	{
		case ZWALL:   DrawCell(pos,b_wall); break;
		case ZLADER:  DrawCell(pos,b_lader); break;
		case ZLADER2: DrawCell(pos,b_lader2); break;
		default:	  DrawCell(pos,0); break;
	}
}

//-- ShowPlayers(): Display the number of current players left
ShowPlayers()
{
	char i;
	DsDisplayBlockClear(114, 108, 4+NPLAYERS*9, 8);
	for(i=0; i<players; i++)
		DsDisplayBitmapDraw(116+i*9, 108, b_manl, 0);
}

//-- GoDown(): Handle the move down on the lader
GoDown(pos) int pos; 
{
	int x,y;
	x = 4 + (pos % 20)*8;
	y = 4 + (pos / 20)*8;

	if(screen[pos+20]==ZLADER)
	{
		onlader=1;
		curpos=pos+40;
		DsDisplayBitmapDraw(x,y+14, (manstate==b_manl)?b_manlc:b_manrc, 0);
		return;
	}
	if(onlader)
	{
		onlader=0;
		RestoreCell(pos-20);
		curpos=pos;
	}
	DrawCell(pos,manstate);
}

//-- Restore2Cells(): restore a cell and the one below
Restore2Cells(pos) int pos;
{
	RestoreCell(pos);
	RestoreCell(pos-20);
}

//-- IsHeart(): Handle the pick up of the hearts by the player
IsHeart()
{
    if(screen[curpos]==ZHEART)
	{
  		score+=POINT_HEART;
		DrawScore();
		hearts--;
		screen[curpos]=0;
	}
}

//-- AnimatePlayer(): Move a player to (x,y) for a short time
AnimatePlayer(x,y,pos) int x; int y; int pos;
{
	DsDisplayBitmapDraw(x,y,manstate,0);
	DsSleep(JUMP_SPEED);
	DsDisplayBlockClear(x,y,8,8);
	Restore2Cells(pos);
}

//-- Movelaser(): Move all the lasers (one cell to the left)
Movelaser()
{
    int x,n;
	if(laserpos>=0)
	   for(n=12; n>=1; n-=2)
			RestoreCell(laserpos+20*n);
	laserpos++;
	if(laserpos==20)
		laserpos=0;
	for(n=12; n>=1; n-=2)
	{
		x = laserpos+20*n;
		if(screen[x])
			DrawCell(x, b_laser);
	}
}

//-- GoUp(): Handle the jumping and the climbing moves
GoUp(pos) int pos;
{
	int dx,dy;
	int x,y;
	x = 4 + (pos % 20)*8;
	y = 4 + (pos / 20)*8;

	if(onlader)
	{
		onlader=0;
		Restore2Cells(pos);
		if(curpos-40>0) curpos=pos-40;
		else if(hearts==0)
		{
			//level completed, go to next level
			score+=POINT_LEVEL;
			curlevel++;
			if(curlevel>=NLEVELS)
				curlevel=0;
			DrawPuzz(1);
			curpos=plevel[0]+20*11;
			manstate=b_manl;
			DrawCell(curpos,manstate);
			onlader=0;
			laserspeed = (maxspeed -= INCREASE_SPEED);
			Movelaser();
			for(x=0; x<12; x++)
			{
				DsDisplayBlockReverse(193,25,21,16,0);
				DsSleep(BLINK_SPEED);
			}
		}
		DrawCell(curpos,manstate);
		return;
	}
	if(screen[pos]==ZLADER2)
	{
		onlader=1;
		DsDisplayBitmapDraw(x,y-2,(manstate==b_manl)?b_manlc:b_manrc,0);
		return;
	}

	dx = (manstate==b_manr)?1:-1;
	if((x<=4+8 && manstate==b_manl) || (x>=4+18*8 && manstate==b_manr)) 
		dx=0; 

	AnimatePlayer(x+4*dx,y-1,pos);
	AnimatePlayer(x+8*dx,y-2,pos);

	curpos+=dx;
	Movelaser();
	IsHeart();

	AnimatePlayer(x+12*dx,y-2,curpos);
	AnimatePlayer(x+16*dx,y-1,curpos);

	if(dx==0)
		Movelaser();

	curpos+=dx;
	IsHeart();
    DsDisplayBitmapDraw(x+16*dx,y,manstate,0);
}

//-- DrawScore(): Display the current score
DrawScore()
{
	char str[5];
	char i=3;
	int  x=score;

	for(str[4]=0; i>=0; i--)
	{
		str[i] = '0'+(x%10);
		x /= 10;
	}	
	DsPrintf(16,108,32,"score:             players:");
	DsDisplayBlockClear(44,108,35,9);
	DsPrintf(45,108,32,str);
}

//-- DrawPuzz(): Display the current level 
//   next=0 -> start from scratch (with all the players)
//   next=1 -> continue the level (with the next player)
DrawPuzz(next) char next;
{
	int s;
	char str[3];
	unsigned char val;
	int i,j;

	DsDisplayBlockClear(1,1,163,4+13*8);

	laserpos = -1;
	manstate = b_manl;
	if(!next)
	{
		score=0;
		players=NPLAYERS;
	}
	DrawScore();	
	ShowPlayers();

	//display level number
	j = curlevel+1;
	str[0] = ' ';
	if(j>9) 
		str[0] = '0'+j/10;
	str[1] = (j%10)+'0';
	str[2] = 0;
	DsDisplayBlockClear(197,26,14,14);
	DsPrintf(197,28,16,str);

	//get level data pointer
	plevel = &m_levels[curlevel*LVL_DATA_SIZE];

	//get player position
	curpos=plevel[0]+20*11;
	DrawCell(curpos, b_manl);
	onlader=0;

	// fill out the 2 first lines with blanks 
	for(s=0; s<2*20; s++)
		screen[s]=0;
	//get exit position of the level
	i = plevel[1]; 
	screen[i]    = ZLADER2;
	screen[20+i] = ZLADER2;

	//decode all the level position of elements into the screen buffer
	for(i=2; i<2+5*5; i++,s+=4)
	{
		// increase line every 5 bytes
		if(i>2 && !((i-2)%5))
			s+=20;
		// get the data byte
		val = plevel[i];
		// decode every 2 bits of the data byte
		for(j=3; j>=0; j--)
		{
			screen[s+j] = val & 3;
			screen[s+j+20] = (screen[s+j]==ZLADER) ? ZLADER2:0;
			val = val >> 2;
		}
	}

	//draw from the decoded screen matrix
	for(s=0; s<20*12; s++)
	{
		switch(screen[s])
		{
			case ZWALL:	  DrawCell(s, b_wall); break;
			case ZLADER:  DrawCell(s, b_lader); break;
			case ZLADER2: DrawCell(s, b_lader2); break;
		}
	}
	for(s=20*12; s<20*13; s++)
	{
		DrawCell(s, b_wall);
		screen[s]=ZWALL;
	}

	//decode hearts level position into a screen matrix
	hearts=0;
	for(j=0; j<8; j++)
	{
		val = plevel[i+j];
		if(val == 0xff)
			break;
		s = (val & 0x1f) + 20 * ((val >> 5) * 2 + 1);
		screen[s] = ZHEART;
		DrawCell(s,b_heart);
		hearts++;
	}
}

//-- IsPlayerKilled(): Check if player is in a hole or touched by a laser
IsPlayerKilled()
{
	char n;
	if(onlader) // players is always safe on laders
		return; 
	// otherwise if touched by a laser or in a hole players is dead
    if((laserpos == curpos%20) || ((curpos/20)!=11 && !screen[curpos+20]))
	{
		// Erase player and move lasers to be ready to restart
		RestoreCell(curpos);
		while(laserpos!=19)
		  Movelaser();
		// make the dead player blinking
		for(n=0; n<9; n++)
		{
		  players += (n%2) ? 1:-1;
		  ShowPlayers();
		  DsSleep(JUMP_SPEED);
		}
		if(players)	
		{
		  // at least one player left -> restart at this level
		  curpos=plevel[0]+20*11;
		  DrawCell(curpos, b_manl);
		  onlader=0;
		}
		else
		{
		  // no player left -> game over
		  start=0;
		  curlevel=0;
		  DrawPuzz(1);
		  DsDisplayBlock(37, 43, 95, 34, 1);
		  DsDisplayBlockClear(38, 44, 92, 31);
		  DsPrintf(53, 54, 16, "Game Over");
		}
	}
}

//-- MakeMove(): Handle player moves and game commands (next,prev & start)
MakeMove(msg) int msg;
{
  int pos;
  pos = curpos;

  switch(msg) 
  {
    case E_PREV:
	   curlevel--;
       if(curlevel<0) curlevel = NLEVELS-1;
       DrawPuzz(0);
	   start=0;
       return;

    case E_NEXT:
       curlevel++;
       if(curlevel >= NLEVELS) curlevel=0;
       DrawPuzz(0);
	   start=0;
       return;

    case E_RESTART:
       DrawPuzz(0);
	   start=1;
	   maxspeed = laserspeed = LASER_SPEED*difficulty - curlevel*INCREASE_SPEED;
	   laserpos=0;
	   Movelaser();
       return;

    default:
	   // no move allowed if game is not started
	   if(start==0)
			return;

	   if(msg>=E_AUPL)	// if(msg==E_AUPL || msg==E_AUPR)
	   {
		   // jumps on the side or climb
		   manstate = b_manl;
		   if(msg==E_AUPR)
			   manstate = b_manr;
		   msg = E_AUP;
	   }
       pos += msg - 50;	// convert the msg into player's new position

       if(msg==E_ALEFT) 
	   {
		   // left move
		   if(curpos%20==0) return;
		   manstate = b_manl;
	   }
       if(msg==E_ARIGHT) 
	   {
		   // right move
		   if((curpos+1)%20==0) return;
		   manstate = b_manr;
	   }
       break;
  }

  RestoreCell(curpos);
  switch(msg)
  {
	case E_AUP:	  GoUp(curpos); break;
	case E_ADOWN: GoDown(curpos); break;
	default:
	  if(onlader)
	  {
		  onlader=0;
		  RestoreCell(curpos-20);
	  }
	  curpos = pos;
	  DrawCell(curpos, manstate);
	  break;
  }

  IsHeart();  
  IsPlayerKilled();
}

//-- SetBmpButton(): Place a bitmap button with event
SetBmpButton(x,y,bmp,evnt) int x; int y; char *bmp; int evnt;
{
	DsDisplayBitmapDraw(x, y, bmp, 0);
	DsEventAdd(x,y,bmp[0],bmp[1],evnt,1);
}

/*--------------------------
    Start of Runner Addin
  --------------------------*/

main() {
	int flag=1;
	MSG msg;

	// intro message
	DsDisplayBlockClear(0,0,240,120);
	DsDisplayBitmapDraw(22,2,b_runner,0);
	DsDisplayLine( 20,25,200,0,85 );
	DsPrintf(20,35,16,"Runner v1.0 - by L. Debert (c)2003");
	DsPrintf(20,52,32,"Pick up all the hearts, don't fall down and avoid");
	DsPrintf(20,62,32,"all the moving lasers.");
	DsPrintf(80,78,12,"Which difficulty?");
	DsDialogTextButton(20,94,95,20,62,"Easy");
	DsDialogTextButton(122,94,95,20,61,"Normal");
	while(flag)
	{
		DsEventMessageGet(msg);
		switch(msg.message)
		{
		case MSG_DS_COMMAND:
			if(msg.bCode==61) 
			{
				difficulty=1;
				flag=0;
			} else if(msg.bCode==62)
			{
				difficulty=2;
				flag=0;
			}
			break;
		}
	}

	// Display game control buttons & frame
	DsDisplayBlockClear(0, 20, 240, 100);
	DsDisplayBitmapDraw(173,4,b_runner,0);
	DsEventClear();
	SetBmpButton(191,43, b_restrt, E_RESTART);
	SetBmpButton(178,25, b_arrowl, E_PREV);
	SetBmpButton(214,25, b_arrowr, E_NEXT);
	DsDisplayBlock(0,0,240,120,0);
	DsDisplayBlock(193,25,21,16,0);
	DsDisplayLine(168,0,0,120,0xFF);

	// Display all player command buttons
	SetBmpButton(178,81, b_arrowl, E_ALEFT);
	SetBmpButton(214,81, b_arrowr, E_ARIGHT);
	SetBmpButton(196,63, b_arrowu, E_AUP);
	SetBmpButton(213,63, b_arrowur, E_AUPR);
	SetBmpButton(178,63, b_arrowul, E_AUPL);
	SetBmpButton(196,97, b_arrowd, E_ADOWN);

	// Display level 1
	start=0;
	curlevel=0;
	DrawPuzz(0);

	flag=1;
	while(flag) 
	{
		DsReadMessage(msg,1);
		switch (msg.message) 
		{
		case MSG_DS_CLOSE:
			flag=0;
			break;

		case MSG_DS_KEY_DOWN:
			if(msg.sCode & 3) flag=0;
			break;

		case MSG_DS_PAINT:
			break;

		case MSG_DS_COMMAND:
			MakeMove(msg.bCode);
			break;
		}
		// control of laser speed & move
	    if(laserpos>=0)
		{
			laserspeed--;
			if(laserspeed==0)
			{
				laserspeed = maxspeed;
				Movelaser();
				IsPlayerKilled();
			}
		}
	}
	DsDisplayBlockClear(0, 0, 240, 120);
	DsAddinTerminate();
}

